<?php

/**
 * @file
 * Contains the dmeEngine class.
 */

/**
 * Class containing the Drupal Markup Engine class.  This class handles filtering markup.
 *
 */
class dmeEngine {
  var $_tag_list = array();
  var $_cid = -1;
  var $_type = '';

  function dmeEngine($tag_array = array()) {
    if (count($tag_array) == 0) {
      $tag_array = dmeEngine::dme_get_full_tag_list();
    }
    $this->_tag_list = $tag_array;
  }

  function set_node($node) {
    $this->_type = 'node';
    $this->_cid = $node->nid;
  }

  /**
	 * Looks through a block of text to determine if context - a node or other id - can be read from it.
	 *
	 * @return bool Is context established?
	 */
  function dme_set_context($text = NULL) {
    if (empty($text)) {
      return (isset($this->_type) && !empty($this->_type) && isset($this->_cid) && !empty($this->_cid));
    }
    else {
      $context_pattern = '!<dme:context type=(["\'])(.*?)\1 cid=(["\'])(.*?)\3 />!i';
      if (preg_match($context_pattern, $text, $matches)) {
        $this->_type = $matches[2];
        $this->_cid = (int)$matches[4];
      }
      return (is_numeric($this->_cid) && in_array($this->_type, array('node', 'comment', 'user')));
    }
  }

  /**
   * Fetches and returns an array of tags which the DME is handling.
   *
   * @return array
   *   Array describing tags.
   */
  function dme_get_full_tag_list() {
    $tags = array();

    $tags[] = array('tag' => 'testpage',
          'allowed_attributes' => 'none',
          'module' => 'dme',
          'type' => 'function',
          'name' => 'print_pagebreak',
      );

    $modules = module_implements('dme_tags');

    foreach ($modules as $module) {
      $module_tags = module_invoke($module, 'dme_tags', 'list');
      foreach ($module_tags as $tag) {
        $tag['module'] = $module;
        $tags[] = $tag;
      }
    }

    return $tags;
  }

  /**
	 * Since the description is displayed as part of a ul, we separate each
	 * description so it looks like multiple items.
	 */
  function get_description() {
    $descriptions = array();
    foreach ($this->_tag_list as $tag) {
      if (isset($tag['description']) && $tag['description'] != '') {
        $descriptions[] = $tag['description'];
      }
    }

    return implode('</li><li>', $descriptions);
  }

  /**
   * This function polls known sources of tags to make and return a list of known tags separated by their type.
   *
   * 'simple' tags are tags which do not enclose any other text, and end in />.
   * 'closed' tags are closed by a similar tag that starts with </
   * For backward compatibility, if we do not find the 'closed' field in the
   * attribute description, we put the tag in both containers.
   *
   * @return array
   *   Tags to prepare for.
   */
  function _dme_get_tag_list($full_tag_list) {

    $result = array('simple' => array(), 'closed' => array());

    foreach ($full_tag_list as $tag) {
      if (!array_key_exists('closed', $tag)) {
        $result['closed'][] = $tag['tag'];
        $result['simple'][] = $tag['tag'];
      }
      else if ($tag['closed']) {
        $result['closed'][] = $tag['tag'];
      }
      else {
        $result['simple'][] = $tag['tag'];
      }
    }
    return $result;
  }

  /**
   * This function takes a text, possibly containing tags in the format <dme:xxxxx >, and a list of possible tags, and converts the angle brackets <> to square [] only for those tags.
   *
   * @param string $text
   *   Text to prepare
   * @param array $tag_list
   *   Array of strings: Tags to convert.
   * @return string
   *   The changed text.
   */
  function _dme_prepare_text($text) {
    $tag_list = $this->_dme_get_tag_list($this->_tag_list);
    if (empty($tag_list['simple']) && empty($tag_list['closed'])) {
      return $text;
    }
    $local_tag_list = $tag_list;

    if (!empty($tag_list['simple'])) {
      // First we test for simple - one piece - tags like <dme:image />
      $regex = "@<dme:(". implode('|', $tag_list['simple']) .')\s*((?:\s+\w+=([-+]?[0-9.]+|(["\']).+?\4))*)\s*/>@si';
      $text = preg_replace($regex, "[dme:$1$2/]", $text);
    }

    if (!empty($tag_list['closed'])) {
      // Next we test for complex - enclosing - tags like <dme:test>blah blah blah</dme:test>
      $regex_with_close_tag = "@<dme:(". implode('|', $tag_list['closed']) .')\s*((?:\s+\w+=([-+]?[0-9.]+|(["\']).+\4))*)\s*>(.*)</dme:\1>@isU';
      $count = 1;
      while ($count) {
        $text = preg_replace($regex_with_close_tag, "[dme:$1$2]$5[/dme:$1]", $text, 1);
        $count = preg_match($regex_with_close_tag, $text);
      }
    }
    return $text;
  }

  /**
   * Processes text, finding tags and turning them into their results.
   *
   * @param string $text
   *  Text to process
   * @param $tag_list
   *  List of tags to handle.
   * @return string
   *  The updated text.
   */
  function _dme_process_text($text) {
    if (count($this->_tag_list) == 0) {
      return $text;
    }

    $tag_name_list = $this->_dme_get_tag_list($this->_tag_list);

    if (!empty($tag_name_list['simple'])) {
      // group 1 is the tag name, group 2 is the parameters, group 3 is the value, but shouldn't be used in the callback,
      //  and group 4 is the type of quote used, if any.
      $regex = "@\[dme:(". implode('|', $tag_name_list['simple']) .')\s*((?:\s+\w+=([-+]?[0-9.]+|(["\']).+?\4))*)\s*/\]@is';
      $text = preg_replace_callback($regex, array($this, "_dme_process_callback"), $text);
    }

    if (!empty($tag_name_list['closed'])) {
      // group 1 is the tag name, group 2 is the parameters, group 3 is the value, but shouldn't be used in the callback,
      //  and group 4 is the type of quote used, if any.  In this regex, group 5 is the inner text of the regex.
      $regex = "@\[dme:(". implode('|', $tag_name_list['closed']) .')\s*((?:\s+\w+=([-+]?[0-9.]+|(["\'])[^\]]+\4))*)\s*\](?!.*\[dme:\1.*\[/dme:\1\])(.*)\[/dme:\1\]@isU';

      $count = 1; // to start us off with.
      while ($count) {
        $text = preg_replace_callback($regex, array($this, "_dme_process_callback"), $text);
        $count = preg_match($regex, $text);
      }
    }

    return $text;
  }

  /**
   * This is called from the preg_replace_callback and calls each tag's handler.
   *
   * @param array $matches
   *  Array of matches/tags from text.
   * @return string
   *  Text to replace the tags with.
   */
  function _dme_process_callback($matches) {
    $nid = 0;
    if (isset($this->_type) && isset($this->_cid)) {
      if ($this->_type == 'node') {
        $nid = $this->_cid;
      }
    }

    // TODO: Change this to work via the array index of the tag list.
    foreach ($this->_tag_list as $tag) {
      if ($tag['tag'] == $matches[1]) {
        $inner_text = (isset($matches[5])?$matches[5]:'');
        $arguments = array();
        $tag_arguments = $this->_dme_process_arguments($matches[2], $tag['allowed_attributes'], $matches[1]);

        if ($inner_text != '') {
          $arguments['dme_inner_text'] = $inner_text;
        }
        $arguments = array_merge($arguments, $tag_arguments);
        if ($tag['type'] == 'function') {
          if (!isset($arguments['nid'])) {
            $arguments['nid'] = $nid;
          }
          if ($tag['module'] == 'dme') {
            return call_user_func_array(array($this, $tag['name']), $arguments);
          }
          else {
            return call_user_func_array($tag['name'], $arguments);
          }
        }
        elseif ($tag['type'] == 'hook') {
          return module_invoke($tag['module'], 'dme_tags', 'call', $tag['name'], $nid, $arguments);
        }
        elseif ($tag['type'] == 'node') {
          // In case of a node type, we're assuming the needed information has been passed in as part of the arguments - node id, whatever.
          // So, for right now, we're not sending the $nid in.
          $tag_details = array($tag['name']);
          if ($arguments['nid']) {
            $node = node_load($arguments['nid']);
            $tag_details[] = $node;
          }
          $arguments = array_merge($tag_details, $arguments);
          return call_user_func_array('theme', $arguments);
        }
      }
    }
    return '';
  }

  /**
   * This function grabs the string of arguments from inside of a tag, and a list of attribute names that are allowed, and returns an array of the valid arguments.
   *
   * @param string $arguments
   *   String of argument/value pairs, like 'nid=27 side="left"'
   * @param mixed $allowed
   *   Either the value 'none' (returns empty string), 'all' (returns all attributes) or an array of allowed attribute names.
   * @return array
   *   Array of allowed attribute/value pairs.
   */
  function _dme_process_arguments($arguments, $allowed, $tag) {
    if ($allowed == 'none' || trim($arguments) == '') {
      return array();
    }
    $result = array();
    $matches = array();
    $regex = '@(\w+)=([-+]?[0-9.]+|(["\'])(.+?)\3)@i';
    while (1 == preg_match($regex, $arguments, $matches)) {
      // Do stuff with matches.
      $matches[1] = strtolower($matches[1]);
      if ($allowed != 'all' && !is_array($allowed)) {
        drupal_set_message("Bad Setup for tag '". $tag ."'.  'allowed_attributes' must be an array if not 'all' or 'none'.", 'error');
        return array();
      }
      if ($allowed == 'all' || in_array($matches[1], $allowed)) {
        if (isset($matches[4]) && $matches[4]) {
          $result[$matches[1]] = $matches[4];
        }
        else {
          $result[$matches[1]] = $matches[2];
        }
      }
      $arguments = preg_replace($regex, '', $arguments, 1);
    }
    return $result;
  }

  function print_pagebreak() {
    return '<!--pagebreak-->';
  }
}